/*
 * Cat decoder example program
 * Version 0.1.1 3FEB2001
 * Copyright 2000, Joe Sunday <sunday@csh.rit.edu>. All Rights Reserved
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE. 
 *
 *  Changes:
 *  0.1.1 
 *    + Fixed error where 'C' characters in some barcode formats xor'ed to
 *      0 and decode loop dropped out prematurely. (aharry at dcomptech.com)
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>

#define BITMASK(b) (0x80 >> ((b) % 8))
#define BITSLOT(b) ((b) / 8)
#define BITSET(a, b) ((a)[BITSLOT(b)] |= BITMASK(b))
#define BITTEST(a, b) ((a)[BITSLOT(b)] & BITMASK(b))

/*
 * Read a string in from the keyboard. Return value is dynamically
 * allocated with malloc() and must be free()d when done.
*/
char* readstring();

/*
 * Decode string from the scanner output into a ascii string. Return
 * value is dynamically allocated with malloc and must be free()d.
*/
char* decode( char* string );

/*
 * Convert c from ascii to the 0-63 values used in the decoding.
*/
char tr( char c );

int main( int argc, char** argv ) {
    char* string;
    char* type;
    char* barcode;
    char* dtype;
    char* dbarcode;
    char* p;
    
    string = readstring();
    p = string;
    while( *p && ( *p != '.' ) ) p++;
    if( *p != '.' ) {
        printf( "Unexpected EOL found\n" );
        exit( 1 );
    }

    p++;
    while( *p && ( *p != '.' ) ) p++;
    if( *p != '.' ) {
        printf( "Unexpected EOL found\n" );
        exit( 1 );
    }
    
    *(p++) = 0;
    type = p;
    while( *p && *p != '.' ) p++;
    if( *p != '.' ) {
        printf( "Unexpected EOL found\n" );
        exit( 1 );
    }
    
    *(p++) = 0;
    barcode = p;
    while( *p && *p != '.' ) p++;
    if( *p != '.' ) {
        printf( "Unexpected EOL found\n" );
        exit( 1 );
    }
    *p = 0;
    dtype = decode( type );
	printf( "%s ", dtype );
    dbarcode = decode( barcode );
    printf( "%s\n", dbarcode );
	free(dtype);
	free(dbarcode);
    free( string );
    return 0;
}

char *readstring( ) {
    char *string;    
    int c;
    int len = 0;
    int maxlen = 50;

    string = (char*) malloc( maxlen );
    c = getc( stdin );                     
    while( (c != EOF) && ( c != '\r' ) &&  ( c != '\n' ) ) {
        string[len] = c;                                          
        len++;
        string[len] = 0;
        if( len == (maxlen - 2) ) {
            char* t = (char*)malloc( maxlen + 50);
            maxlen += 50;
            strncpy( t, string, len );           
            free( string );
            string = t;              
         }
         c = getc( stdin );
     }             
     return string;
}

char *decode( char* string )
{
	unsigned char  *work1, *work2;
	int i, len1, len2, pos1=0, pos2=0;

	len1 = strlen(string);
	len2 = (len1 * 6) / 8;
	work1 = (unsigned char*)calloc( 1, len1+1);
	work2 = (unsigned char*)calloc( 1, len2+1);


	for( i=0; i < len1; i++)
		work1[i]=tr(string[i]);

	len1 = strlen(string) * 8;

	while( pos1 < len1 )
	{
		/* Skip two bits */
		pos1 += 2;

		/* Copy six bits to work buffer 2 */
		for(i = 0; i <6; i++, pos1++, pos2++ )
			if(BITTEST(work1,pos1))
				BITSET(work2,pos2);
	}

	/* XOR each byte of the output with 'C' */
	for(i=0; i < len2; i++)
		work2[i] ^= 'C';

	free(work1);
	return work2;
}

char tr( char c ) {
    char out;
    
    if( (c >= 'a') && (c <= 'z') ) {
        out = c - 'a';
    }
    else if( ( c >= 'A' ) && ( c <= 'Z' ) ) {
        out = c - 'A' + 26;
    }
    else if( ( c >= '0' ) && ( c <= '9' ) ) {
        out = c - '0' + 52;
    }
    else if( c == '+' ) {
        out = 62;
    }
    else if( c == '-' ) {
        out = 63;
    }
    else {
        printf( "Bad character /%c/ in code.\n", c );
    }
    return out;
}

